package zrazumovskiy;

import java.util.*;

public class RungeKutta
{
	private ArrayList xList = new ArrayList();
	private ArrayList yList = new ArrayList();

	private double xInit, yInit;
	private double x0, xN, y0, yN;
	private double xScale, yScale;

	private FunctionVx parserF;
	private FunctionVy parserG;
	
	private Parser parser;
	private DrawPanel panel; 

	private double redrawDeltaT;
	
	public RungeKutta(double x0, double xN, double y0, double yN, FunctionVx parserF, FunctionVy parserG)
	{
		this.x0 = x0;
		this.xN = xN;
		this.y0 = y0;
		this.yN = yN;

		this.parserF = parserF;
		this.parserG = parserG;
	}


	public RungeKutta(double x0, double xN, double y0, double yN, FunctionVx parserF, FunctionVy parserG, Parser
	parser, DrawPanel panel, double redrawDeltaT)
	{
		this.x0 = x0;
		this.xN = xN;
		this.y0 = y0;
		this.yN = yN;

		this.parserF = parserF;
		this.parserG = parserG;

		this.panel = panel;
		this.parser = parser;
		this.redrawDeltaT =  redrawDeltaT;
	}

	private double getLocalVector()
	{
		double Vx = parserF.getValue();
		double Vy = parserG.getValue();

		return (Math.sqrt(Vx*Vx + Vy*Vy));
	}

	private double selectIntegrStep(double integrKoeff)
	{
		double h;
		double vector = this.getLocalVector();
		if( vector > 1)
			h = integrKoeff / vector / 5.;
		else
		{
			if( integrKoeff > 0 )
				h = vector / 5.;
			else
				h = -vector / 5. ;
		}

		return h;
	}

	private boolean checkLoopValues(double xCurr, double yCurr, double h, int i, double vector)
	{
		if( (xCurr < x0) || (xCurr > xN) || (yCurr < y0) || (yCurr > yN) )
			return false;

		if( (Math.abs(h) < 1E-8 )|| (vector<1e-8))
			return false;

		return true;
	}

	private boolean checkLoopValues(double xCurr, double yCurr, double h, int i, double tCurr, double tFinal, double vector)
	{
		if( Double.isInfinite(tFinal) )
		{
			if( (xCurr < x0) || (xCurr > xN) || (yCurr < y0) || (yCurr > yN) )
				return false;

			if( (Math.abs(h) < 1E-8) || (vector< 1e-8))
				return false;

		}
		else
		{
                    
			if( tCurr >= tFinal)
				return false;

		}

		return true;
	}

	private void setParserValues(double x, double y, double t)
	{
		parserF.addVariable("x", x);
		parserF.addVariable("y", y);
		parserF.addVariable("t", t);
		parserG.addVariable("x", x);
		parserG.addVariable("y", y);
		parserG.addVariable("t", t);
	}
	private void setParserValues(double x, double y)
	{
		parserF.addVariable("x", x);
		parserF.addVariable("y", y);
		parserG.addVariable("x", x);
		parserG.addVariable("y", y);
	}

	private double[] takeStep(double xCurr, double yCurr, double h)
	/*
	Given the x_n and y_n and step size h calculates x_{n+1} and y_{n+1} and estimates of the truncation
	errors for both x and y. Note! This procedure does not advance time!
	*/
	{
		double xtmp, ytmp, xtmpStar, ytmpStar;
		double [] result = new double [4];
		// Cash-Karp Parameters
		double [] a = new double[6];
		a[0]=0;
		a[1]=0.2;
		a[3]=0.6;
		a[4]=1.;
		a[5]=0.875;
		double [] c = new double[6];
		c[0]=37./378.;
		c[1]=0;
		c[2]=250./621.;
		c[3]=125./594.;
		c[4]=0;
		c[5]=512./1771.;
		double [] cStar = new double[6];
		cStar[0]=2825./27648.;
		cStar[1]=0;
		cStar[2]=18575./48384.;
		cStar[3]=13525./55296.;
		cStar[4]=277./14336.;
		cStar[5]=0.25;
		double [][] b = new double[6][6];
		b[1][0]=0.2;
		b[2][0]=3./40.;
		b[3][0]=0.3;
		b[4][0]=-11./54.;
		b[5][0]=1631./55296.;
		b[2][1]=9./40.;
		b[3][1]=-0.9;
		b[4][1]=5./2.;
		b[5][1]=175./512.;
		b[3][2]=6./5.;
		b[4][2]=-70./27.;
		b[5][2]=575./13824.;
		b[4][3]=35./27.;
		b[5][3]=44275./110592.;
		b[5][4]=253./4096.;
		// parameters k and l for x and y steps
		double [] k = new double [6];
                double [] l = new double [6];

		// load the current point. if the time is advanced, add current time here

		setParserValues(xCurr, yCurr);

		// k1's (actually, k(0)
		k[0] = h * parserF.getValue();
		l[0] = h * parserG.getValue();

		// now calculate k2..k6
		for (int i=1; i<6; i++)
		{
			/* if the time is advanced, the time point is t_n+a(i)*h
			where t_n is the time at the beginning of the step.
			*/
			xtmp=xCurr;
			ytmp=yCurr;

			for (int j=0; j<i; j++)
			{
				xtmp=xtmp+b[i][j]*k[j];
				ytmp=ytmp+b[i][j]*l[j];
			}

			setParserValues(xtmp, ytmp);

			k[i] = h * parserF.getValue();
			l[i] = h * parserG.getValue();
		}

		// ok, having calculated k's we are ready to get the next point
		// This is O(h^5) estimate
		xtmp=xCurr;
		ytmp=yCurr;

		for (int i=0; i<6;i++)
		{
			xtmp=xtmp+c[i]*k[i];
			ytmp=ytmp+c[i]*l[i];
		}
		// This is O(h^4) estimate

		xtmpStar=xCurr;
		ytmpStar=yCurr;

		for (int i=0; i<6;i++)
		{
			xtmpStar=xtmpStar+cStar[i]*k[i];
			ytmpStar=ytmpStar+cStar[i]*l[i];
		}
		/* return the results. the first element is x, the second - y, the third -
		estimate of the
		truncation error in x, the last one - the same thing for y
		*/
		result [0]=xtmp;
		result [1]=ytmp;
		result [2]=Math.abs(xtmp-xtmpStar);
		result [3]=Math.abs(ytmp-ytmpStar);

		return result;
	}

	private double[] takeStep(double xCurr, double yCurr, double tCurr, double h)
	/*
	Given the x_n and y_n and step size h calculates x_{n+1} and y_{n+1} and
	estimates of the truncation
	errors for both x and y. 
	*/
	{
		double xtmp, ytmp, xtmpStar, ytmpStar, ttmp;
		double [] result = new double [4];
		// Cash-Karp Parameters
		double [] a = new double[6];
		a[0]=0;
		a[1]=0.2;
		a[3]=0.6;
		a[4]=1.;
		a[5]=0.875;
		double [] c = new double[6];
		c[0]=37./378.;
		c[1]=0;
		c[2]=250./621.;
		c[3]=125./594.;
		c[4]=0;
		c[5]=512./1771.;
		double [] cStar = new double[6];
		cStar[0]=2825./27648.;
		cStar[1]=0;
		cStar[2]=18575./48384.;
		cStar[3]=13525./55296.;
		cStar[4]=277./14336.;
		cStar[5]=0.25;
		double [][] b = new double[6][6];
		b[1][0]=0.2;
		b[2][0]=3./40.;
		b[3][0]=0.3;
		b[4][0]=-11./54.;
		b[5][0]=1631./55296.;
		b[2][1]=9./40.;
		b[3][1]=-0.9;
		b[4][1]=5./2.;
		b[5][1]=175./512.;
		b[3][2]=6./5.;
		b[4][2]=-70./27.;
		b[5][2]=575./13824.;
		b[4][3]=35./27.;
		b[5][3]=44275./110592.;
		b[5][4]=253./4096.;
		// parameters k and l for x and y steps
		double [] k = new double [6];
                double [] l = new double [6];

		// load the current point. if the time is advanced, add current time here

		setParserValues(xCurr, yCurr, tCurr);

		// k1's (actually, k(0)
		k[0] = h * parserF.getValue();
		l[0] = h * parserG.getValue();

		// now calculate k2..k6
		for (int i=1; i<6; i++)
		{
			/* if the time is advanced, the time point is t_n+a(i)*h
			where t_n is the time at the beginning of the step.
			*/
			xtmp = xCurr;
			ytmp = yCurr;
			ttmp = tCurr + a [i] * h;
			for (int j=0; j<i; j++)
			{
				xtmp=xtmp+b[i][j]*k[j];
				ytmp=ytmp+b[i][j]*l[j];
			}

			setParserValues(xtmp, ytmp, ttmp);

			k[i] = h * parserF.getValue();
			l[i] = h * parserG.getValue();
		}

		// ok, having calculated k's we are ready to get the next point
		// This is O(h^5) estimate
		xtmp=xCurr;
		ytmp=yCurr;

		for (int i=0; i<6;i++)
		{
			xtmp=xtmp+c[i]*k[i];
			ytmp=ytmp+c[i]*l[i];
		}
		// This is O(h^4) estimate

		xtmpStar=xCurr;
		ytmpStar=yCurr;

		for (int i=0; i<6;i++)
		{
			xtmpStar=xtmpStar+cStar[i]*k[i];
			ytmpStar=ytmpStar+cStar[i]*l[i];
		}
		/* return the results. the first element is x, the second - y, the third -
		estimate of the
		truncation error in x, the last one - the same thing for y
		*/
		result [0]=xtmp;
		result [1]=ytmp;
		result [2]=Math.abs(xtmp-xtmpStar);
		result [3]=Math.abs(ytmp-ytmpStar);

		return result;
	}

	private double getNewStepSize(double [] stepResult,  double delta0x, double delta0y, double h, double power)
	// Given the desired accuracies delta0x , delta0y and achieved accuracy taken from stepResult, calculates
	// the new step size 
	{
		double xh, yh;
		xh=0.9*h*Math.exp(Math.log(delta0x/stepResult[2])*power);
		yh=0.9*h*Math.exp(Math.log(delta0y/stepResult[3])*power);	
		if (Math.abs(xh)<Math.abs(yh)) yh=xh;
		if (h>0)
                {
                  if (yh > h) {yh=Math.min(yh, 5*h);}
                         else {yh=Math.max(yh, 0.1*h);}
                }
                else
                {
                  if (yh < h) {yh=Math.max(yh, 5*h);}
                         else {yh=Math.min(yh, 0.1*h);}

                }
                return yh;
	}


	private void iterateRungeKutta(double integrKoeff)
	/*
	Implements 4th order Runge-Kutta-Cash-Karp method with adaptive step size
	control.
	Note! This procedure does not advance time.
	*/
	{
		boolean b = true;
		double epsilon=1e-8; //relative error
		double tiny=1e-30; //some small number
		double h = 0;
                double delta0x, delta0y; // desired accuracy
                double vxCurr, vyCurr; // current values of v used in scaling relations

		int i = xList.size();

		// the first point
		double xCurr = xInit;
		double yCurr = yInit;
		double vector;

		// results of each step are returned using this array
		double [] stepResult = new double [4];


		setParserValues(xCurr, yCurr);

		// initial guess for the step size
		vector = this.getLocalVector();
		h = integrKoeff / vector;

		do
		{
			xList.add(new Double(xCurr));
			yList.add(new Double(yCurr));

			i++;
			// attempt a step
			stepResult=takeStep(xCurr, yCurr,h);

			// now estimate the accuracy that we want
        		setParserValues(xCurr, yCurr);

                        vxCurr=parserF.getValue();
                        vyCurr=parserG.getValue();

                        delta0x=epsilon*(Math.abs(xCurr)+Math.abs(h*vxCurr)+tiny);
                        delta0y=epsilon*(Math.abs(yCurr)+Math.abs(h*vyCurr)+tiny);

			// Decide whether the step was successful
			if ((stepResult[2]< delta0x) &&
				(stepResult[3] < delta0y))
			{
			// the step succeeded. accept x,y and calculate h for the next step
			xCurr = stepResult[0];
			yCurr = stepResult[1];
			h=getNewStepSize(stepResult, delta0x, delta0y ,h, 0.2);
			}
                        else
                        {
			// the step failed. calculate new h and repeat the step
			h=getNewStepSize(stepResult, delta0x, delta0y,h, 0.25);
                        stepResult=takeStep(xCurr,yCurr,h);
			xCurr = stepResult[0];
			yCurr = stepResult[1];
                        }
                        vector = Math.sqrt(vxCurr*vxCurr+vyCurr*vyCurr);
			b = checkLoopValues(xCurr, yCurr, h, i, vector);
		}
		while (b);
		xList.add(new Double(xCurr));
		yList.add(new Double(yCurr));

	}

	private void iterateRungeKutta(double t0, double integrKoeff, double tFinal, boolean redraw)
	{
		double tCurr = t0;

		boolean b = true;
		double epsilon=1e-8; //relative error
		double tiny=1e-30; //some small number
		double h = 0;
                double delta0x, delta0y; // desired accuracy
                double vxCurr, vyCurr; // current values of v used in scaling relations

		int i = xList.size();

		// the first point
		double xCurr = xInit;
		double yCurr = yInit;
		double vector;

		// results of each step are returned using this array
		double [] stepResult = new double [4];
		ArrayList tList = new ArrayList();


		setParserValues(xCurr, yCurr, tCurr);

		// initial guess for the step size
		vector = this.getLocalVector();
		h = integrKoeff / vector;

		int redrawStep = 0;

		do
		{
			xList.add(new Double(xCurr));
			yList.add(new Double(yCurr));
			tList.add(new Double(tCurr));

			if (redraw) 
			{
				if (Math.floor((tCurr-t0)/redrawDeltaT) > redrawStep)
				{
					redrawStep = (int) Math.floor((tCurr-t0)/redrawDeltaT);
					redrawVectorField(tCurr);
                                        drawAnimation(xList, yList);
				}
			
			}
			
			i++;
			
			// attempt a step
                        
			if (h>0)
			{
			   	if (tCurr+h> tFinal)
			      		h = tFinal - tCurr + 1e-30; 
			}      
			else
			{
				if (tCurr + h < tFinal)
					h = tFinal - tCurr - 1e-30;
			}                       
			stepResult=takeStep(xCurr, yCurr, tCurr, h);

			// now estimate the accuracy that we want
        		setParserValues(xCurr, yCurr, tCurr);

                        vxCurr=parserF.getValue();
                        vyCurr=parserG.getValue();

                        delta0x=epsilon*(Math.abs(xCurr)+Math.abs(h*vxCurr)+tiny);
                        delta0y=epsilon*(Math.abs(yCurr)+Math.abs(h*vyCurr)+tiny);

			// Decide whether the step was successful
			if ((stepResult[2]< delta0x) &&
				(stepResult[3] < delta0y))
			{
			// the step succeeded. accept x,y and calculate h for the next step
			xCurr = stepResult[0];
			yCurr = stepResult[1];
			tCurr = tCurr + h;
			h=getNewStepSize(stepResult, delta0x, delta0y, h, 0.2);
			}
                        else
                        {
			// the step failed. calculate new h and repeat the step
			h=getNewStepSize(stepResult, delta0x, delta0y,h, 0.25);
                        stepResult=takeStep(xCurr,yCurr, tCurr, h);
			xCurr = stepResult[0];
			yCurr = stepResult[1];
			tCurr = tCurr + h;
                        }
                        vector = Math.sqrt(vxCurr*vxCurr+vyCurr*vyCurr);
			if( integrKoeff > 0 )
				b = checkLoopValues(xCurr, yCurr, h, i, tCurr, tFinal,vector);
			else
				b = checkLoopValues(xCurr, yCurr, h, i, tFinal, tCurr, vector);
		}
		while( b );
		
		{
		xList.add(new Double(xCurr));
		yList.add(new Double(yCurr));
		tList.add(new Double(tCurr));
		}
	}

	public ArrayList[] calculateStreamLines(double xInit, double yInit, double integrKoeff, double xScale, double yScale)
	{
		this.xInit = xInit;  
		this.yInit = yInit;  
		this.xScale = xScale;  
		this.yScale = yScale;  
		
		xList.clear();
		yList.clear();

		iterateRungeKutta(integrKoeff);
		ArrayList xTemp = new ArrayList(xList);
		ArrayList yTemp = new ArrayList(yList);

		xList.clear();
		yList.clear();

		iterateRungeKutta(-integrKoeff);
		xList = Utils.reverseArrayList(xList);
		yList = Utils.reverseArrayList(yList);

		xList.addAll(xTemp);
		yList.addAll(yTemp);

		ArrayList[] result = {xList, yList};

		return result;
	}

	public ArrayList[] calculatePathLines(double xInit, double yInit, double tCurr, double tFinal, 
	    double integrKoeff, double xScale, double yScale, boolean redraw)
	{
		this.xInit = xInit;
		this.yInit = yInit;
		this.xScale = xScale;
		this.yScale = yScale;

		xList.clear();
		yList.clear();

		iterateRungeKutta(tCurr, integrKoeff, tFinal, redraw);

		ArrayList[] result = {xList, yList};

		return result;
	}

	public ArrayList[] calculateStreakLines(double xInit, double yInit, double
tFinal, double tStart, double integrKoeff, int m, double xScale, double yScale)
	{
		ArrayList[] array = new ArrayList[2];

		double tCurr = tFinal;
		double vector;
		boolean b;
		double h;
		int i = 0;
                
                double prevxPoint,prevyPoint, stepMag;
		ArrayList x = new ArrayList();
		ArrayList y = new ArrayList();

		setParserValues(xInit, yInit, 0.0);
                prevxPoint = xInit;
                prevyPoint = yInit;
                
                int redrawStep = 0;
                
		do
		{
			i++;
			array = calculatePathLines(xInit, yInit, tCurr, tFinal, integrKoeff, xScale,
yScale, false);
			h = selectIntegrStep(integrKoeff);
			tCurr = tCurr - h;
			int lastPointIndex = array[0].size() - 1;
			Double xLastPoint = (Double)array[0].get(lastPointIndex);
			Double yLastPoint = (Double)array[1].get(lastPointIndex);
			x.add(xLastPoint);
			y.add(yLastPoint);
// animation
			if (Math.floor((tFinal-tCurr)/redrawDeltaT) > redrawStep)
			{
                            redrawStep = (int) Math.floor((tFinal-tCurr)/redrawDeltaT);
                            redrawVectorField(tCurr);
                            drawAnimation(x, y);
			}
		                        
                        
//end of animation                        
                        
                        b = checkLoopValues( xLastPoint.doubleValue(), yLastPoint.doubleValue(), h, i,1
);                     
                        stepMag = Math.sqrt((xLastPoint.doubleValue()-prevxPoint)*(xLastPoint.doubleValue()-prevxPoint)+
                           (yLastPoint.doubleValue()-prevyPoint)*(yLastPoint.doubleValue()-prevyPoint));
                        prevxPoint = xLastPoint.doubleValue();
                        prevyPoint = yLastPoint.doubleValue();
		}
		while( b & (! ((stepMag<1e-6) & (i>1))  ));

		ArrayList[] result = {x, y};

		return result;
	}

	public ArrayList[] calculateContourPoints(ArrayList xCoord, ArrayList yCoord,
double t0, double tFinal, double integrKoeff)
	{
		int length = xCoord.size();
		ArrayList[] array = new ArrayList[2];
		ArrayList x = new ArrayList();
		ArrayList y = new ArrayList();
		double xInit, yInit;

		for( int i = 0; i < length; i++ )
		{
			xInit = ((Double)xCoord.get(i)).doubleValue();
			yInit = ((Double)yCoord.get(i)).doubleValue();
                        
                        if( t0 < tFinal )
        			array = calculatePathLines(xInit, yInit, t0, tFinal, integrKoeff, xScale,
yScale, false);
                        else
        			array = calculatePathLines(xInit, yInit, t0, tFinal, -integrKoeff, xScale,
yScale, false);
			int lastPointIndex = array[0].size() - 1;
			Double xLastPoint = (Double)array[0].get(lastPointIndex);
			Double yLastPoint = (Double)array[1].get(lastPointIndex);
			x.add(xLastPoint);
			y.add(yLastPoint);
		}

		ArrayList[] result = {x, y};

		return result;
	}
	private void drawAnimation(ArrayList xList, ArrayList yList )
        {
            ArrayList [] singleLineCoords = {xList, yList};
            panel.drawAnimationLines(singleLineCoords);
        }
	private void redrawVectorField(double t)
	{
		parser.setTime(t);
		double[][] Vx = null;
		double[][] Vy = null;
		try
		{
			Vx = parser.calculateF();
			Vy = parser.calculateG();
		}
		catch(UndefinedValueException ex)
		{}
		
		panel.setV(Vx, Vy);
                panel.normalizeVectorField();
		panel.paintImmediately(0, 0, panel.getWidth(), panel.getHeight());
		
	}
}



 	
